# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020, 2022       Universit'e de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#
from starpu import starpupy
import starpu
import inspect

def access(**kwargs):
	def access_decorator(func):
		# write access modes in f.access attribute
		setattr(func,'starpu_access', kwargs)
		return func
	return access_decorator

# set mode as **kwargs of func
def set_access(func, **kwargs):
	# write access modes in f.access attribute
	setattr(func,'starpu_access', kwargs)
	return func


















